/*  File:  tile.c  */

#include <ctype.h>
#include <stdarg.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#define PLUS '@'
#define MIN  '.'

int
    **a,
    *f,
    x_n = 0,
    y_n = 0,
    all = 0,
    swap = 0;

char
    **s,
    *programname,
    *no_mem_buffer,
    Out_of_memory [] = "Out of memory!";

void
    show (void),
    get_programname (char const *argv0),
    process_args (int argc, char *argv []),
    syntax (void),
    errit (char const *format, ...),
    *s_malloc (size_t size);
char
    *get_arg (int *argc, char *argv [], int *index);
int
    solution (void);

int main (int argc, char *argv [])
{
    int
	i,
        x,
        y;

    no_mem_buffer = (char *) malloc (1024);

    get_programname (argv [0]);

    process_args (argc, argv);

    if (y_n > x_n) {
        i = x_n;
        x_n = y_n;
        y_n = i;
        swap = 1;
    }

    a = (int **) s_malloc ((x_n + 2) * sizeof (int *));
    for (x = 0; x <= x_n + 1; x++)
        a [x] = (int *) s_malloc ((y_n + 2) * sizeof (int));

    f = (int *) s_malloc ((y_n + 2) * sizeof (int));

    s = (char **) s_malloc (y_n * sizeof (char *));
    for (y = 0; y < y_n; y++) {
        s [y] = (char *) s_malloc ((x_n + 1) * sizeof (char));
        s [y][x_n] = '\0';
    }

    printf ("\nTesting symmetric\n");
    memset (f + 1, 0, y_n * sizeof (int));
    for (y = 1; y; ) {

	if (solution ())
            show ();

	for (y = (y_n + 1) / 2; y; --y) {
	    if (y_n + 1 - y != y)
                f [y_n + 1 - y] ^= 1;
	    if (f [y] ^= 1)
		break;
	}
    }

    printf ("\nTesting others\n");
    memset (f + 1, 0, y_n * sizeof (int));
    for (y = 1; y; ) {

        i = 0;
        for (y = y_n / 2; y >= 1; y--)
            if (f [y] != f [y_n + 1 - y]) {
                i = 1;
                break;
	    }

        if (i)
	    if (solution ())
                show ();

	for (y = y_n; y; --y)
	    if (f [y] ^= 1)
		break;
    }

    return 0;
}

int solution ()
{
    register int
	x,
	y;

    for (x = x_n; x; x--)
        for (y = y_n; y; y--)
	    a [x][y] = 1;

    for (y = y_n; y; y--)
	if (f [y]) {
	    a [1][y] ^= 1;
	    a [1][y - 1] ^= 1;
	    a [1][y + 1] ^= 1;
	    a [2][y] ^= 1;
	}

    for (x = 2; x <= x_n; x++)
        for (y = y_n; y; y--)
	    if (a [x - 1][y]) {
		a [x][y] ^= 1;
		a [x][y - 1] ^= 1;
		a [x][y + 1] ^= 1;
		a [x + 1][y] ^= 1;
	    }

    for (y = y_n; y; y--)
	if (a [x_n][y])
	    return 0;

    return 1;
}

void show ()
{
    register int
	x,
	y;

    for (x = x_n; x; x--)
        for (y = y_n; y; y--)
	    a [x][y] = 1;

    for (y = y_n; y; y--)
	if (f [y]) {
	    a [1][y] ^= 1;
	    a [1][y - 1] ^= 1;
	    a [1][y + 1] ^= 1;
	    a [2][y] ^= 1;
            s [y - 1][0] = PLUS;
	} else
            s [y - 1][0] = MIN;

    for (x = 2; x <= x_n; x++) {
        for (y = y_n; y; y--)
	    if (a [x - 1][y]) {
		a [x][y] ^= 1;
		a [x][y - 1] ^= 1;
		a [x][y + 1] ^= 1;
		a [x + 1][y] ^= 1;
                s [y - 1][x - 1] = PLUS;
	    } else
                s [y - 1][x - 1] = MIN;
    }

    putchar ('\n');
    if (swap) {
        for (x = 0; x < x_n; x++) {
            for (y = 0; y < y_n; y++)
                putchar (s [y][x]);
            putchar ('\n');
        }
    } else
        for (y = 0; y < y_n; y++)
            puts (s [y]);

    if (! all)
        exit (0);
}

void syntax ()
{
    fprintf (
	stderr,
	"\nTiling Problem Solver\n\n"
        "(c) P. Kleiweg 1996\n\n"
	"Syntax: %s -x int [-y int] [-a]\n\n"
        "-x: width\n"
        "-y: height\n"
        "-a: all\n\n",
	programname
    );
    exit (1);
}

void process_args (int argc, char *argv [])
{
    int
	index;

    if (argc == 1)
        syntax ();

    for (index = 1; index < argc; index++) {
	if (argv [index][0] == '-') {
	    switch (argv [index][1]) {
		case 'x':
		    x_n = atoi (get_arg (&argc, argv, &index));
                    if (x_n < 1)
                        errit ("Illegal value for width");
		    break;
		case 'y':
		    y_n = atoi (get_arg (&argc, argv, &index));
                    if (y_n < 1)
                        errit ("Illegal value for height");
		    break;
		case 'a':
		    all = 1;
		    break;
		default:
                    errit ("Invalid option '%s'", argv [index]);
	    }
	} else
            errit ("Invalid parameter \"%s\"", argv [index]);
    }
    if (! x_n)
        errit ("Width not defined");
    if (! y_n)
        y_n = x_n;
}

char *get_arg (int *argc, char *argv [], int *index)
{
    int
	i;
    i = 1;
    while (argv [*index][++i]) {
	if (! isspace (argv [*index][i]))
	    return argv [*index] + i;
    }
    if (*index == *argc - 1)
	errit ("Missing argument for \"%s\"", argv [*index]);
    return argv [++*index];
}

void get_programname (char const *argv0)
{
    char
	*p;
    p = strrchr (argv0, '/');
    if (p)
	programname = strdup (p + 1);
    else
	programname = strdup (argv0);
}

void *s_malloc (size_t size)
{
    void
	*p;

    p = malloc (size);
    if (! p) {
	free (no_mem_buffer);
	errit (Out_of_memory);
    }
    return p;
}

void errit (char const *format, ...)
{
    va_list
	list;

    fprintf (stderr, "\nError %s: ", programname);

    va_start (list, format);
    vfprintf (stderr, format, list);

    fprintf (stderr, "\n\n");

    exit (1);
}
